import enum

from pyttman.core.containers import MessageMixin


class PyttmanPluginIntercept(enum.StrEnum):
    """
    Enum for the different intercepts available for
    Pyttman plugins to hook into. The intercepts are
    called at different stages of the application
    runtime, allowing for custom workflows to be
    implemented.
    """
    before_app_start = "before_app_start"
    after_app_stops = "after_app_stops"
    before_router = "before_router"
    before_intent = "before_intent"
    after_intent = "after_intent"
    before_entity_extraction = "before_entity_extraction"
    after_entity_extraction = "after_entity_extraction"
    no_intent_match = "no_intent_match"

class PyttmanPlugin:
    """
    The PluginBase class offers an API for contributors to
    develop custom workflows which interact with different
    parts of a Pyttman application runtime.

    Implement methods as desired to interact with the
    various layers.

    The instance is long-lived throughout the application runtime.

    To control when the plugin is allowed to run, provide a list
    of allowed intercepts in the constructor. If no intercepts are
    provided, the plugin will never run.

    allowed_intercepts: list[PyttmanPluginIntercepts]
        A list of allowed points of intercept, where the plugin is allowed
        to intercept the application runtime. An empty list will allow
        no intercepts, effectively disabling the plugin.
    """
    def __init__(self,
                 allowed_intercepts: list[PyttmanPluginIntercept] or None = None,
                 app = None):
        self.allowed_intercepts = set(allowed_intercepts) if allowed_intercepts else set()
        self.app = app

    def on_app_start(self):
        """
        Execute code when the app starts, with the 'app' object available.
        """
        pass


    def allowed_to_intercept_at(self, intercept: PyttmanPluginIntercept):
        """
        Check if the plugin is allowed to run for the given intercept.
        """
        return intercept in self.allowed_intercepts

    def before_app_start(self, app):
        """
        Implement this method to have code execute before the app
        starts.
        """
        return app

    def after_app_stops(self, app):
        """
        Implement this method to have code execute when the
        app has exited
        """
        return app

    def before_router(self, message: MessageMixin):
        """
        Implement this method to interact with the processing
        of a message, before it's processed by the Router.
        This hook is useful to manipulate the message object
        before it's passed to the Router, affecting which
        intent it will match.
        """
        return message

    def before_intent(self, message: MessageMixin):
        """
        Implement this method to interact with the processing
        of a message, before it's processed by the matching Intent.
        """
        return message

    def after_intent(self, reply: MessageMixin):
        """
        Implement this method to interact with the Reply object
        from an Intent, before it's delivered to the Client for
        a platform response.
        """
        return reply

    def before_entity_extraction(self, message: MessageMixin):
        """
        Implement this method to interact with the processing
        of a message, before entities are parsed from the message.
        This is one way to affect the entities that are extracted
        from the message, when it's been passed to the matching intent.
        """
        return message

    def after_entity_extraction(self, message: MessageMixin):
        """
        This hook allows you to interfere with the message object
        (with associated entities on message.entities), to manipulate
        or change any entities extracted from the message before
        it's passed to the Intent.
        """
        return message

    def no_intent_match(self, message: MessageMixin):
        """
        Implement this method to interact with the message
        when no intent matches the message.
        """
        return message



import traceback

import pyttman
from pyttman.core import exceptions
from pyttman.core.containers import Message
from pyttman.core.plugins.base import PyttmanPlugin, PyttmanPluginIntercept

if __name__ != "__main__":
    try:
        import certifi
        import mongoengine
    except ImportError:
        message = (
            "\nMongoEnginePlugin: In order to use the Mongo Engine "
            "Plugin with Pyttman, you must install mongoengine and "
            "certifi.\n\n <pip install mongoengine, certifi>\n")
        print(message)
        exit(0)


class MongoEnginePlugin(PyttmanPlugin):
    """
    The MongoEnginePlugin offers a one-stop object for a
    Pyttman application to work with MongoEngine -
    a comprehensive ORM developed by the team at MongoDB.

    An application-wide connection is configured, making the use of
    mongoengine simple.

    When using this plugin, you can start develop Models and start
    using them in your application right away.

    If 'user_table_name' is provided, the 'message.author' object
    will be set as the matching User object in the mongodb database,
    matching the original author reference.

    In order to use this plugin, 'mongoengine' from pip, must be
    installed in the application.

    Tip! This plugin is accessible as `app.loaded_plugins.MongoEnginePlugin`
    in your application.
    """

    class MessageUserBinding:
        """
        Configure how the MongoEnginePlugin can prepare Message objects
        by setting a matching User as the `message.user` attribute (if any),
        on messages as they're routed to Intent classes.

        Provide the model class you use in your application for users,
        and which column to use when trying to identify a matching user
        based on the primary ID in the foreign user class*.

        * By foreign user*, we mean the message.author object which is
          different depending on which platform you are connecting your
          Pyttman app with. In Discord for example, it's an Author object.
          In this case, the Author.id property is used to try and match
          with your model.
        """
        user_model_class: type = None
        """ Define the Model class for Users in the application, i.e. MyUserModel """
        column_for_matching_user_to_author: str = None
        """ Define which column to match your user model with the 'Message.author' 
            object, i.e. "my_column_name"
        """
        custom_queryset_method_name: str = None
        """ Optionally, you can provide the name of a method on the QuerySet for 
            your model here which can return a matching User object provided a 
            message.author.id 
        """

        def __init__(self,
                     user_model_class: type = None,
                     column_for_matching_user_to_author: str = None,
                     custom_queryset_method_name: str = None):
            if user_model_class is None:
                raise exceptions.PyttmanPluginException(
                    "You must define your user model class in the user binding. "
                    "Example: `user_model_class=MyUserClass`")
            elif column_for_matching_user_to_author and custom_queryset_method_name:
                raise exceptions.PyttmanPluginException(
                    "You have defined an ambiguous configuration for user matching: "
                    "Both an attribute and a method is provided for matching. "
                    "Please choose either to define a column for direct matching "
                    "Author.id, or provide the name of a method in the QuerySet "
                    "for your model to call, when matching. ")

            self.user_model_class = user_model_class
            self.column_for_matching_user_to_author = column_for_matching_user_to_author
            self.custom_queryset_method_name = custom_queryset_method_name

    def __init__(self,
                 db_name: str,
                 host: str,
                 port: int | str,
                 user_binding: MessageUserBinding = None,
                 username: str = None,
                 password: str = None,
                 allowed_intercepts: list[PyttmanPluginIntercept] or None = None):
        self.user_binding = user_binding
        self.db_name = db_name
        self.host = host
        self.port = int(port)
        self.username = username
        self.password = password
        super().__init__(allowed_intercepts)

    def on_app_start(self):
        """
        Set up a connection with MongoDB using mongoengine
        with provided credentials.
        """
        pyttman.logger.log("- [MongoEnginePlugin]: connecting...")
        mongoengine.connect(
            tlsCAFile=certifi.where(),
            db=self.db_name,
            host=self.host,
            username=self.username,
            password=self.password,
            port=self.port)
        self.password = None
        pyttman.logger.log("- [MongoEnginePlugin]: connection successful.")


    def after_app_stops(self, app):
        pyttman.logger.log("- [MongoEnginePlugin]: disconnecting...")
        mongoengine.disconnect_all()
        pyttman.logger.log("- [MongoEnginePlugin]: disconnection successful.")

    def before_intent(self, message: Message):
        if self.user_binding is None:
            return
        model = self.user_binding.user_model_class

        try:
            if custom_method := self.user_binding.custom_queryset_method_name:
                method = getattr(model.objects, custom_method)
                user = method(message.author.id)
            else:
                local_column = self.user_binding.column_for_matching_user_to_author
                query = {f"{local_column}__exact": message.author.id}
                user = model.objects.get(**query)
        except Exception:
            pyttman.logger.log(
                level="error",
                message="MongoEnginePlugin was unable to perform user "
                        f"matching. Author: {message.author}, traceback: "
                        f"{traceback.format_exc()}")
            user = None
        message.user = user
        return message


import json
from dataclasses import dataclass, field
from itertools import zip_longest
from pathlib import Path
from typing import Callable

import requests

import pyttman
from pyttman.core.containers import MessageMixin, Reply
from pyttman.core.plugins.base import PyttmanPlugin, PyttmanPluginIntercept


@dataclass
class OpenAiRequestPayload:
    model: str
    system_prompt: str
    user_prompt: str

    def as_json(self):
        return {
            "model": self.model,
            "messages": [
                {
                    "role": "system",
                    "content": self.system_prompt
                },
                {
                    "role": "user",
                    "content": self.user_prompt
                }
            ]
        }

@dataclass
class RagMemoryBank:
    """
    The OpenAiRagMemoryBank is a dataclass that holds
    the conversation history with a user, and the
    memories that the AI should remember.

    Saving the RAG data can be defined by the user, building the app.
    They can provide callbacks for us to use when CRUD:ing memories.
    """
    file_path: Path | None = None
    memories: dict[str, list[str]] = field(default_factory=dict)
    callbacks: dict[str, Callable or None] = field(default_factory=dict)

    def __post_init__(self):
        self.callbacks = {
            "purge_all_memories": None,
            "purge_memories": None,
            "append_memory": None,
            "get_memories": None,
        }

    def _execute_callback(self, callback_name: str, *args) -> any:
        """
        Execute a callback if it's defined.
        """
        try:
            if (callback := self.callbacks.get(callback_name)) is not None:
                response = callback(*args)
                return True if response is None else response
        except Exception as e:
            pyttman.logger.log(level="error",
                               message=f"OpenAIPlugin: RagMemoryBank: "
                                       f"callback {callback_name} failed: {e}")

    def _load_memories_from_file(self):
        """
        Load the memories source
        """
        if self.file_path is None:
            raise ValueError("OpenAIPlugin: RagMemoryBank: Using file storage "
                             "fallback failed. No file path defined for the "
                             "memory bank.")

        if not self.file_path.exists():
            self.file_path.parent.mkdir(parents=True, exist_ok=True)
            self.save_to_file()

        with open(self.file_path, "r", encoding="utf-8") as f:
            data = json.loads(f.read())
            self.memories = data["memories"]

    def _get_memory_from_file(self, key):
        """
        Fallback, unless user implements callback. Use file-based memories.
        """
        if not self.memories:
            self._load_memories_from_file()
        return self.memories.get(str(key), [])

    def _add_memory_to_file(self, key, memory):
        if not self.memories:
            self._load_memories_from_file()

        key = str(key)
        if self.memories.get(key) is None:
            self.memories[key] = [memory]
        else:
            self.memories[key].append(memory)
        self.save_to_file()

    def purge_all_memories(self):
        """
        Purge all memories.
        """
        if self._execute_callback("purge_all_memories") is not None:
            return
        self.memories.clear()

    def purge_memories(self, key: str):
        """
        Purge all memories for a given key.
        """
        if self._execute_callback("purge_memories", key) is not None:
            return
        self.memories[key] = []

    def get_memories(self, key: str) -> tuple[str] or list[str]:
        """
        Return the memories for a given key.
        """
        if (memories := self._execute_callback("get_memories", key)) is not None:
            if not isinstance(memories, (list, tuple)):
                raise ValueError("OpenAIPlugin: The memories callback must "
                                 "return a list or tuple.")
            return memories
        return self._get_memory_from_file(key)

    def add_memory(self, key: str, memory: str):
        """
        Append a memory to the memory bank.
        """
        if callback_return := self._execute_callback("add_memory", key, memory):
            return callback_return
        self._add_memory_to_file(key, memory)

    def save_to_file(self):
        """
        Save the memories to a file.
        """
        with open(self.file_path, "w", encoding="utf-8") as f:
            data = self.as_json()
            f.write(json.dumps(data, indent=4))

    def as_json(self):
        return {"memories": self.memories}

    def memories_as_str(self, key: str) -> str:
        """
        Return the memories as a string.
        """
        key = str(key)
        base = "These are your long term memories with this user: "
        memories = self.get_memories(key)
        return base + "\n".join(memories)


class OpenAIPlugin(PyttmanPlugin):
    """
    The OpenAIPlugin offers seamless integrations with the OpenAI API.
    Use the plugin to define pre-prompts that can be used to pre- or post
    process your message in your Pyttman application.

    An example is to use a pre-defined system prompt to correct spelling
    mistakes before the message is passed to the intent matching system.

    Another example is to use the GPT as a post-processor, to generate
    a response based on the intent matched by the Pyttman application.

    Or - use the GPT to generate a response from scratch, when no intent
    matches the user's message. This would be a great way to combine the
    rule-based intent matching system with an AI model.

    The plugin supports RAG: conversational mode. When this is enabled, the
    plugin will keep a conversation history for each user in memory, and
    use this history to generate responses. This can be useful to keep
    the conversation flowing naturally, and to keep the context of the
    conversation intact. While recommended, it's important to note that
    the data is stored in memory, non-encrypted, and will be lost when
    the application is restarted.
    """

    conversation_prompt = ("You will get a copy of the conversation history "
                           "with this user so far. Your previous messages "
                           "are prefixed with 'You: '. Do not include this "
                           "'You: ' in your actual replies. Respond according "
                           "to the users' last message, naturally as if conversing "
                           "with a human, taking the history in the dialogue "
                           "you've already had. \n\n")

    detect_memory_prompt = ("Determine if this message contains something the user "
                            "shares with you that you are expected to remember. It "
                            "could be anything from a name, a place, a date, a task, "
                            "or something they share about their life. It could be a "
                            "direct encouragement to remember something for the future, "
                            "or a clear directive to create a memory of something. It "
                            "could also just be a detail shared with you, that a human "
                            "would remember about them. If you think you should remember "
                            "something, Read the content of what to remember from the "
                            "user message and return the memory in this format: "
                            "'[MEMORY]: {your memory content here}'. If the message does "
                            "not match memory making or is a question, return 0")

    def __init__(self,
                 api_key: str,
                 model: str,
                 system_prompt: str = None,
                 max_tokens: int = None,
                 enable_conversations: bool = False,
                 enable_memories: bool = False,
                 max_conversation_length: int = 32_000,
                 allowed_intercepts: list[PyttmanPluginIntercept] = None,
                 purge_all_memories_callback: callable or None = None,
                 purge_memories_callback: callable or None = None,
                 add_memory_callback: callable or None = None,
                 get_memories_callback: callable or None = None):
        super().__init__(allowed_intercepts)
        self.api_key = api_key
        self.model = model
        self.system_prompt = system_prompt
        self.session = requests.Session()
        self.url = "https://api.openai.com/v1/chat/completions"
        self.max_tokens = max_tokens
        self.api_key = api_key
        self.enable_conversations = enable_conversations
        self.max_conversation_length = max_conversation_length
        self.enable_memories = enable_memories
        self.rag_memories_path: Path | None = None
        self.long_term_memory: RagMemoryBank | None = None
        self.conversation_rag = {}

        self._purge_all_memories_callback = purge_all_memories_callback
        self._purge_memories_callback = purge_memories_callback
        self._add_memory_callback = add_memory_callback
        self._get_memories_callback = get_memories_callback

        self.session.headers.update({"Content-Type": "application/json"})
        self.session.headers.update({"Accept-Type": "application/json"})
        self.session.headers.update({"Authorization": f"Bearer {self.api_key}"})

        del self.api_key

    def on_app_start(self):
        if (static_files_dir := self.app.settings.STATIC_FILES_DIR) is None:
            static_files_dir = Path(self.app.settings.APP_BASE_DIR / "static")
        self.rag_memories_path = static_files_dir / "rag_memories" / "memories.json"
        self.long_term_memory = RagMemoryBank(self.rag_memories_path)

        self.long_term_memory.callbacks["purge_all_memories"] = self._purge_all_memories_callback
        self.long_term_memory.callbacks["purge_memories"] = self._purge_memories_callback
        self.long_term_memory.callbacks["add_memory"] = self._add_memory_callback
        self.long_term_memory.callbacks["get_memories"] = self._get_memories_callback

        pyttman.logger.log("- [OpenAIPlugin]: Plugin started.")

    def _prepare_rag_prompt(self, message: MessageMixin) -> str:
        """
        Use RAG to prepend conversation history with this user to
        the outgoing llm request.
        """
        if not self.enable_conversations:
            return message.as_str()

        conversation = ""
        for user_message, ai_message in zip_longest(
            self.conversation_rag[message.author.id]["user"],
            self.conversation_rag[message.author.id]["ai"],
                fillvalue=""
        ):
            if user_message:
                conversation += f"User: {user_message}\n"
            if ai_message:
                conversation += f"You: {ai_message}\n"
        return conversation + f"User: {message.as_str()}\n"

    def _prepare_payload(self, message: MessageMixin) -> dict:
        """
        Prepare a payload towards OpenAI.
        """
        if self.enable_conversations:
            user_prompt = self._prepare_rag_prompt(message)
            system_prompt = self.system_prompt + self.conversation_prompt
            pyttman.logger.log(f" - [OpenAIPlugin]: conversation size "
                               f"for user {message.author.id}: {len(user_prompt)}")
        else:
            system_prompt = self.system_prompt
            user_prompt = message.as_str()

        if self.enable_memories:
            memories = self.long_term_memory.get_memories(message.author.id)
            system_prompt += (f"\nThese are your long term memories "
                              f"with this user: {"\n".join(memories)}")

        return OpenAiRequestPayload(
            model=self.model,
            system_prompt=system_prompt,
            user_prompt=user_prompt).as_json()

    def before_router(self, message: MessageMixin):
        """
        Executes before the router resolves the message to an intent.
        """
        payload = self._prepare_payload(message)
        if self.max_tokens:
            payload["max_tokens"] = self.max_tokens

        try:
            response = self.session.post(self.url, json=payload)
            response_content = response.json()["choices"][0]["message"]["content"]
            message.content = response_content
            return message
        except requests.exceptions.RequestException as e:
            pyttman.logger.log(level="error",
                               message=f"OpenAIPlugin: Request to "
                                       f"OpenAI API failed: {e}")
            return Reply("I'm sorry, I couldn't generate a response for you.")

    def create_memory_if_applicable(self, message) -> str or None:
        """
        Create curated RAG memory prompts, stored to file locally
        """
        payload = OpenAiRequestPayload(
            model=self.model,
            system_prompt=self.detect_memory_prompt,
            user_prompt=message.as_str()).as_json()

        try:
            response = self.session.post(self.url, json=payload)
            memory = response.json()["choices"][0]["message"]["content"]
            if str(memory) == "0":
                return None
            return memory
        except requests.exceptions.RequestException as e:
            pyttman.logger.log(level="error",
                               message=f"OpenAIPlugin: Request to "
                                       f"OpenAI API failed: {e}")
            return None


    def no_intent_match(self, message: MessageMixin) -> Reply | None:
        """
        Hook. Executed when no intent matches the user's message.
        """
        if new_memory := self.create_memory_if_applicable(message):
            self.long_term_memory.add_memory(message.author.id, new_memory)

        if self.conversation_rag.get(message.author.id) is None:
            self.conversation_rag[message.author.id] = {"user": [message.as_str()], "ai": []}
        else:
            self.conversation_rag[message.author.id]["user"].append(message.as_str())

        while True:
            user_length = len("".join(self.conversation_rag[message.author.id]["user"]))
            ai_length = len("".join(self.conversation_rag[message.author.id]["ai"]))

            if user_length + ai_length > self.max_conversation_length:
                self.conversation_rag[message.author.id]["user"].pop(0)
                self.conversation_rag[message.author.id]["ai"].pop(0)
            else:
                break

        error_response = Reply("I'm sorry, I couldn't generate a response for you.")
        payload = self._prepare_payload(message)

        if self.max_tokens:
            payload["max_tokens"] = self.max_tokens

        try:
            response = self.session.post(self.url, json=payload)
        except requests.exceptions.RequestException as e:
            pyttman.logger.log(level="error",
                               message=f"OpenAIPlugin: Request to OpenAI API failed: {e}")
            return error_response

        if not response.ok:
            pyttman.logger.log(level="error",
                               message=f"OpenAIPlugin: Request to OpenAI "
                                       f"API failed: {response.text}")
            return error_response

        try:
            gpt_content = response.json()["choices"][0]["message"]["content"]
            self.conversation_rag[message.author.id]["ai"].append(gpt_content)
            if new_memory:
                gpt_content = f"Memory updated.\n{gpt_content}"
            return Reply(gpt_content)
        except KeyError:
            pyttman.logger.log(level="error",
                               message="OpenAIPlugin: No response from OpenAI API.")
            return error_response




# Set the current working dir to parent directory
import os
import shutil
import subprocess
import sys
from pathlib import Path
from time import sleep

sys.path.append(Path.cwd().parent.as_posix())

if not Path("setup.py").exists():
    print("You're not in the right directory. Run this script from the "
          r"project's root directory, e.g. 'C:\users\your_user\projects\pyttman'.")
    exit(-1)

LAB_ENV_PATH = Path.cwd() / Path("dev_env")
BUILD_OUTPUT_PATH = Path.cwd() / "dist"

if __name__ == "__main__":
    if (LAB_ENV_PATH / "venv").exists():
        shutil.rmtree((LAB_ENV_PATH / "venv").as_posix())

    if not Path("dist").exists():
        subprocess.check_call("python devtools/build.py".split())

    LAB_ENV_PATH.mkdir(exist_ok=True)
    os.chdir(LAB_ENV_PATH.as_posix())
    subprocess.run("python -m virtualenv venv".split())

    while not Path("venv").exists():
        sleep(0.01)

    package_file = [
        i for i in os.listdir(BUILD_OUTPUT_PATH.as_posix()) if i.endswith("tar.gz")
    ].pop()

    print(f"Installing package: '{package_file}' from your local build...")
    package_file = (BUILD_OUTPUT_PATH / package_file).as_posix()
    venv_python = (LAB_ENV_PATH / "venv/scripts/python.exe").as_posix()
    subprocess.run(f"{venv_python} -m pip install multidict".split())
    subprocess.run(f"{venv_python} -m pip install {package_file}".split())

    print("\nFinished! Here's how to get started:",
          f"1. Activate the virtual environment:\n\tcd dev_env\n\tvenv/scripts/activate",
          f"2. Run the command 'pyttman' to see available commands to the Pyttman CLI",
          "3. If it's the first time you're running Pyttman, run 'pyttman new app {app_name}' to create a new project."
          "4. Run 'pyttman dev {app_name}' to start the development server.",
          "5. If you've made changes to the Pyttman framework which you want to test in your testing project, "
          "run this script again. Your app will be left untouched, but the Pyttman version is upgraded to "
          "your current HEAD in the Pyttman repo.",
          sep="\n")



import inspect
from typing import Tuple

from pyttman.core.intent import Intent
from pyttman.core.internals import _generate_name
from pyttman.core.mixins import PrettyReprMixin
from pyttman.core.storage.basestorage import Storage


class Ability(PrettyReprMixin):

    """
    Base class for an Ability.

    The Ability is an encapsulating class which
    holds Intent subclasses in its 'intents' tuple.

    It provides an encapsulating scope for Intents
    which shares Storage object. Since this data may
    be sensitive and irrelevant for other functionality
    in the app, this encapsulation provides comfort and
    security for the Intent endpoints not to access or
    destroy data which they're not meant to.

    The Ability class can be configured when used, to
    set up objects in the Storage object, or any other
    code that needs to run before the app starts - can
    be put in or called by the 'configure' method
    """
    description = "Unavailable"
    intents: Tuple = None

    __repr_fields__ = ("name",)

    def __init__(self, **kwargs):
        super().__init__()

        self.storage = Storage()
        self.name = _generate_name(self.__class__.__name__)
        self.before_create()
        [setattr(self, k, v) for k, v in kwargs.items()]

        if self.intents is not None:
            self.__validate_intents()

    def __validate_intents(self):
        """
        Assert that the tuple contains references to
        'Intent' subclasses and nothing else.
        """
        try:
            iter(self.intents)
            if not isinstance(self.intents, Tuple):
                raise TypeError
        except TypeError:
            raise TypeError(f"The 'intents' property must be tuple, "
                            f"got {type(self.intents)}'.")
        else:
            for intent_class in self.intents:
                if not inspect.isclass(intent_class) or not \
                        issubclass(intent_class, Intent):
                    raise TypeError(
                        f"Intent '{intent_class}' is not defined correctly. "
                        f"Intents must be class references, and must inherit "
                        "from the 'Intent' base class.\nBe sure to only "
                        "mention the name of the class, and not instantiate "
                        "it when defining the 'intents' property in your "
                        "Ability class.\nHint: Change '(FooIntent(), "
                        "BarIntent())' to '(FooIntent, BarIntent).")

    def before_create(self):
        """
        Lifecycle hook. This hook method is executed before the
        ability is created, when the application starts, but it
        does have access to the Storage object.
        """
        pass



import re
from typing import Union

from pyttman.core.containers import MessageMixin
from pyttman.core.entity_parsing.entity import Entity


class Identifier:
    """
    Base class for an Identifier.

    The Identifier class aids in identifying
    strings based on their patterns, as in
    if they contain integers, special characters,
    whether they are capitalized, have a date-like
    format to them, and alike.

    Subclasses can define in a very granular way, how
    a string can look by strict rules, or less
    so with less defined criterion.

    The Identifier will use regex to assess
    the similarity of given string.

    This class is subclassed and patterns are
    defined in the tuple 'patterns' as raw
    python strings (prepend the string with 'r').
    The regex pattern is evaluated by the EntityFieldValueParser
    at runtime.
    """
    patterns = (r"^.*$",)
    min_length = None
    max_length = None
    start_index = 0

    def __init__(self, **kwargs):
        try:
            self.patterns = tuple([re.compile(pat) for pat in self.patterns])
        except Exception as e:
            raise AttributeError("Identifier pattern could not compile") from e
        [setattr(self, k, v) for k, v in kwargs.items()]

    def __repr__(self):
        return f"{self.__class__.__name__}(patterns={self.patterns})"

    def _assert_length_requirement(self, value: str) -> bool:
        """
        Assert the element identified complies
        with min_length and max_length fields
        :param value: str, value to me examined
        :return: bool, complies with both min_value and
                 max_value or not. if min_value and/or
                 max_value is omitted, they are considered
                 compliant in all situations, bypassing any
                 length of strings.
        """
        return ((len(value) > self.min_length
                 if self.min_length is not None else True) and
                (len(value) < self.max_length
                 if self.max_length is not None else True))

    def try_identify_entity(self, message: MessageMixin) -> Union[Entity,
                                                                  None]:
        """
        Evaluates if any element in the content of
        a Message object matches with its pattern.

        :return Entity: Element in message.content which matched
                     the pattern assigned, or None if none found
        """
        for pattern in self.patterns:
            try:
                for i, elem in enumerate(message.content[self.start_index:]):
                    if re.match(pattern, elem) and self\
                            ._assert_length_requirement(elem):
                        return Entity(value=elem,
                                      index_in_message=self.start_index + i)
            except IndexError:
                return None
        return None


class CellPhoneNumberIdentifier(Identifier):
    """ Identifies whether a string is similar to a cell number """
    patterns = (r"^(\d{3}.\d{4}.\d{3})|(\d{10})|(\d{3}.\d{3}.\d{4})$",)


class DateTimeStringIdentifier(Identifier):
    """ Identifies whether a string looks like a date string """
    patterns = (r"^(\d{4}).(\d{2})",
                r"^(\d{2}).(\d{2}).(\d{2})",
                r"^(\d{4}).(\d{2}).(\d{2}).(\d{2}):(\d).*$",
                r"^(\d{2}).(\d{2}).(\d{2}).(\d{2}):(\d{2}).*$")


class DateTimeFormatIdentifier(Identifier):
    """ identifies a datetime format configuration string """
    patterns = (r"^%.*$",)


class IntegerIdentifier(Identifier):
    """ identifies all integers """
    patterns = (r"[0-9]+",)


class CapitalizedIdentifier(Identifier):
    """
    This identifier class will target strings which
    are capitalized, meaning they begin with a capital
    letter.
    """
    patterns = (r"\b[A-Z][a-zA-Z]*\b",)


NumberIdentifier = IntegerIdentifier
FloatIdentifier = IntegerIdentifier
